module Exam.Part3.Specification where

open import Data.Nat using (ℕ; zero; suc)
open import Relation.Binary.PropositionalEquality using (_≡_; refl)

open import Exam.Util.Fin using (Fin; zero; suc)

-- Fixities --------------------------------------------------------------------

infix  4 _↪_ _↪*_
infix  4 _⊢_⦂_
infixl 5 _,_
infix  5 `λ_
infixl 7 _·_
infix  9 `_
infixl 9 _[_]

-- Axioms ----------------------------------------------------------------------

-- Functional Extensionality
postulate
  fun-ext : ∀ {ℓ₁ ℓ₂} {A : Set ℓ₁} {B : A → Set ℓ₂} {f g : (x : A) → B x}
    → (∀ x → f x ≡ g x)
    → f ≡ g

-- No other postulates are allowed!

-- Syntax ----------------------------------------------------------------------

data Term : ℕ → Set where
  `_      : ∀ {n} → Fin n → Term n
  `λ_     : ∀ {n} → Term (suc n) → Term n
  _·_     : ∀ {n} → Term n → Term n → Term n
  ∀[x⦂_]_ : ∀ {n} → Term n → Term (suc n) → Term n
  `Set    : ∀ {n} → Term n

-- Substitution ----------------------------------------------------------------

-- Renamings

Ren : ℕ → ℕ → Set
Ren m n = Fin m → Fin n

extᵣ : ∀ {m n} → Ren m n → Ren (suc m) (suc n)
extᵣ ρ zero    = zero
extᵣ ρ (suc x) = suc (ρ x)

ren : ∀ {m n}
  → Ren m n
  → (Term m → Term n)
ren ρ (` x)                = ` (ρ x)
ren ρ (`λ e)               = `λ (ren (extᵣ ρ) e)
ren ρ (e₁ · e₂)            = (ren ρ e₁) · (ren ρ e₂)
ren ρ (∀[x⦂ e₁ ] e₂)       = ∀[x⦂ ren ρ e₁ ] ren (extᵣ ρ) e₂
ren ρ `Set                 = `Set

wk : ∀ {n} → Ren n (suc n)
wk x = suc x

-- Substitutions

Sub : ℕ → ℕ → Set
Sub m n = Fin m → Term n

extₛ : ∀ {m n} → Sub m n → Sub (suc m) (suc n)
extₛ σ zero    = ` zero
extₛ σ (suc x) = ren wk (σ x)

sub : ∀ {m n}
  → Sub m n
  → (Term m → Term n)
sub σ (` x)                = σ x
sub σ (`λ e)               = `λ (sub (extₛ σ) e)
sub σ (e₁ · e₂)            = (sub σ e₁) · (sub σ e₂)
sub σ (∀[x⦂ e₁ ] e₂)       = ∀[x⦂ sub σ e₁ ] sub (extₛ σ) e₂
sub σ `Set                 = `Set

0↦ : ∀ {n} → Term n → Sub (suc n) n
(0↦ e) zero    = e
(0↦ e) (suc x) = ` x

_[_] : ∀ {n} → Term (suc n) → Term n → Term n
e₁ [ e₂ ] = sub (0↦ e₂) e₁

-- Semantics -------------------------------------------------------------------

-- Values

mutual
  data Value : ∀ {n} → Term n → Set where
    val-λ : ∀ {n} {e : Term (suc n)}
      → Value e
        ------------
      → Value (`λ e)
    val-∀ : ∀ {n} {e₁ : Term n} {e₂ : Term (suc n)}
      → Value e₁
      → Value e₂
        --------------------
      → Value (∀[x⦂ e₁ ] e₂)
    val-Set : ∀ {n}
      → Value {n = n} `Set
    val-neu : ∀ {n} {e : Term n}
      → Neutral e
        ---------
      → Value e

  data Neutral : ∀ {n} → Term n → Set where
    neu-` : ∀ {n} {x : Fin n}
      → Neutral (` x)
    neu-· : ∀ {n} {e₁ e₂ : Term n}
      → Neutral e₁
      → Value e₂
        -----------------
      → Neutral (e₁ · e₂)

-- Reduction

data _↪_ : ∀ {n} → Term n → Term n → Set where
  β-λ : ∀ {n} {e₁ : Term (suc n)} {e₂ : Term n}
    → (`λ e₁) · e₂ ↪ e₁ [ e₂ ]

  ξ-λ : ∀ {n} {e e' : Term (suc n)}
    → e ↪ e'
    ---------------
    → `λ e ↪ `λ e'

  ξ-·₁ : ∀ {n} {e₁ e₂ e₁' : Term n}
    → e₁ ↪ e₁'
    ---------------------
    → e₁ · e₂ ↪ e₁' · e₂ 

  ξ-·₂ : ∀ {n} {e₁ e₂ e₂' : Term n}
    → e₂ ↪ e₂'
    ---------------------
    → e₁ · e₂ ↪ e₁ · e₂'

  ξ-∀₁ : ∀ {n} {t₁ t₁' : Term n} {t₂ : Term (suc n)}
    → t₁ ↪ t₁'
    -------------------------------
    → ∀[x⦂ t₁ ] t₂ ↪ ∀[x⦂ t₁' ] t₂

  ξ-∀₂ : ∀ {n} {t₁ : Term n} {t₂ t₂' : Term (suc n)}
    → t₂ ↪ t₂'
    -------------------------------
    → ∀[x⦂ t₁ ] t₂ ↪ ∀[x⦂ t₁ ] t₂'

-- Reflexive, Transitive Closure of Reduction

data _↪*_ : ∀ {n} → Term n → Term n → Set where
  ↪*-refl : ∀ {n} {e : Term n}
    → e ↪* e
  ↪*-step : ∀ {n} {e₁ e₂ e₃ : Term n}
    → e₁ ↪  e₂
    → e₂ ↪* e₃
    → e₁ ↪* e₃

↪*-trans : ∀ {n} {e₁ e₂ e₃ : Term n}
  → e₁ ↪* e₂
  → e₂ ↪* e₃
  → e₁ ↪* e₃
↪*-trans ↪*-refl q = q
↪*-trans (↪*-step x p) q = ↪*-step x (↪*-trans p q)

module ↪*-Reasoning where
  infix 1 begin_
  infixr 2 _↪⟨_⟩_ _↪*⟨_⟩_ _≡⟨_⟩_ _≡⟨⟩_
  infix 3 _∎

  begin_ : ∀ {n} {e₁ e₂ : Term n} → e₁ ↪* e₂ → e₁ ↪* e₂
  begin p = p

  _↪⟨_⟩_ : ∀ {n} (e₁ {e₂} {e₃} : Term n) → e₁ ↪ e₂ → e₂ ↪* e₃ → e₁ ↪* e₃
  _ ↪⟨ p ⟩ q = ↪*-step p q

  _↪*⟨_⟩_ : ∀ {n} (e₁ {e₂} {e₃} : Term n) → e₁ ↪* e₂ → e₂ ↪* e₃ → e₁ ↪* e₃
  _ ↪*⟨ p ⟩ q = ↪*-trans p q

  _≡⟨_⟩_ : ∀ {n} (e₁ {e₂} {e₃} : Term n) → e₁ ≡ e₂ → e₂ ↪* e₃ → e₁ ↪* e₃
  _ ≡⟨ refl ⟩ q = q

  _≡⟨⟩_ : ∀ {n} (e₁ {e₂} {e₃} : Term n) → e₁ ↪* e₂ → e₁ ↪* e₂
  _ ≡⟨⟩ q = q

  _∎ : ∀ {n} (e : Term n) → e ↪* e
  _ ∎ = ↪*-refl

-- Conversion

-- Two terms are convertible iff they reduce to a common term.

infix 4 _≈_
record _≈_ {n} (e₁ e₂ : Term n) : Set where
  constructor mk-≈
  field
    e-common     : Term n
    e₁↪*e-common : e₁ ↪* e-common
    e₂↪*e-common : e₂ ↪* e-common

open _≈_ public

-- Typing ----------------------------------------------------------------------

data Context : ℕ → Set where
  ∅   : Context zero
  _,_ : ∀ {n} → Context n → Term n → Context (suc n)

lookup : ∀ {n} → Context n → Fin n → Term n
lookup (Γ , e) zero    = ren wk e
lookup (Γ , e) (suc x) = ren wk (lookup Γ x)

data _⊢_⦂_ : ∀ {n} → Context n → Term n → Term n → Set where

  ⊢-` : ∀ {n t} {Γ : Context n} (x : Fin n)
    → lookup Γ x ≡ t
      --------------
    → Γ ⊢ ` x ⦂ t

  ⊢-λ : ∀ {n} {Γ : Context n} {e : Term (suc n)} {t₁ : Term n} {t₂ : Term (suc n)}
    → Γ ⊢ t₁ ⦂ `Set
    → Γ , t₁ ⊢ e ⦂ t₂
      -----------------------
    → Γ ⊢ `λ e ⦂ ∀[x⦂ t₁ ] t₂

  ⊢-· : ∀ {n} {Γ : Context n} {e₁ e₂ : Term n} {t₁ : Term n} {t₂ : Term (suc n)}
    → Γ ⊢ e₁ ⦂ ∀[x⦂ t₁ ] t₂
    → Γ ⊢ e₂ ⦂ t₁ 
      -----------------------
    → Γ ⊢ e₁ · e₂ ⦂ t₂ [ e₂ ]

  ⊢-∀ : ∀ {n} {Γ : Context n} {t₁ : Term n} {t₂ : Term (suc n)}
    → Γ ⊢ t₁ ⦂ `Set
    → Γ , t₁ ⊢ t₂ ⦂ `Set
      -----------------------
    → Γ ⊢ ∀[x⦂ t₁ ] t₂ ⦂ `Set

  ⊢-Set : ∀ {n} {Γ : Context n}
    → Γ ⊢ `Set ⦂ `Set

  ⊢-≈ : ∀ {n} {Γ : Context n} {e : Term n} {t t' : Term n}
    → t ≈ t'
    → Γ ⊢ e ⦂ t
    ------------
    → Γ ⊢ e ⦂ t'
